package com.code.aaron.speech;

import com.code.aaron.speech.SpeechService;

import android.media.AudioFormat;
import android.media.AudioRecord;
import android.media.MediaRecorder;

import com.google.api.gax.core.CredentialsProvider;
import com.google.api.gax.core.FixedCredentialsProvider;
import com.google.api.gax.rpc.ClientStream;
import com.google.api.gax.rpc.ResponseObserver;
import com.google.api.gax.rpc.StreamController;
import com.google.auth.oauth2.ServiceAccountCredentials;
import com.google.cloud.speech.v1.RecognitionConfig;
import com.google.cloud.speech.v1.SpeechClient;
import com.google.cloud.speech.v1.SpeechSettings;
import com.google.cloud.speech.v1.StreamingRecognitionConfig;
import com.google.cloud.speech.v1.StreamingRecognizeRequest;
import com.google.cloud.speech.v1.StreamingRecognizeResponse;
import com.google.protobuf.ByteString;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import io.flutter.plugin.common.EventChannel;
import io.flutter.plugin.common.PluginRegistry.Registrar;

//import io.grpc.okhttp.OkHttpChannelProvider;
//import io.grpc.netty.NettyChannelProvider;

/** SpeechPlugin */
public class SpeechPlugin implements EventChannel.StreamHandler {
    private static final String CHANNEL_NAME = "aaron.code.com/speech";

    /** Plugin registration. */
    public static void registerWith(Registrar registrar) {
    final EventChannel channel = new EventChannel(registrar.messenger(), CHANNEL_NAME);
    channel.setStreamHandler(new SpeechPlugin());
    }

    private EventChannel.EventSink eventSink;

    // Audio recorder + initial values
    private static volatile AudioRecord recorder;

    // Google auth
    public static final List<String> SCOPE =
            Collections.singletonList("https://www.googleapis.com/auth/cloud-platform");
    private static final String HOSTNAME = "speech.googleapis.com";
    private static final int PORT = 443;
    private static String credentialsPath;

    private int AUDIO_SOURCE = MediaRecorder.AudioSource.DEFAULT;
    private int SAMPLE_RATE = 16000;
    private int CHANNEL_CONFIG = AudioFormat.CHANNEL_IN_MONO;
    private int AUDIO_FORMAT = AudioFormat.ENCODING_PCM_16BIT;
    private int BUFFER_SIZE = AudioRecord.getMinBufferSize(SAMPLE_RATE, CHANNEL_CONFIG, AUDIO_FORMAT);

    // Runnable management
    private volatile boolean record = false;
    private volatile boolean isRecording = false;

    @Override
    public void onListen(Object args, EventChannel.EventSink eventSink) {
        if (isRecording) return;

        ArrayList<String> config = (ArrayList<String>) args;
        switch (config.size()) {
            case 0:
                eventSink.error("-21", "TooFewArguments", null);
            case 1:
                credentialsPath = config.get(0);
                break;
            default:
                eventSink.error("-22", "TooManyArguments", null);
        }

        this.eventSink = eventSink;

        // Try to initialize and start the recorder
        recorder = new AudioRecord(AUDIO_SOURCE, SAMPLE_RATE, CHANNEL_CONFIG, AUDIO_FORMAT, BUFFER_SIZE);
        if (recorder.getState() != AudioRecord.STATE_INITIALIZED) eventSink.error("-1", "PlatformError", null);
        recorder.startRecording();

        // Start runnable
        record = true;
        new Thread(startStreaming(credentialsPath)).start();
    }

    @Override
    public void onCancel(Object args) {
        // Stop runnable
        record = false;

        // Stop and reset audio recorder
        if (recorder != null) {
            recorder.stop();
            recorder.release();
            recorder = null;
        }
    }

    private Runnable startStreaming(String credentialsPath) {
        return new Runnable() {
            @Override
            public void run() {
                try {
                    File credentialsFile = new File(credentialsPath);
                    //if (!isFile(credentialsFile)) eventSink.error("-25", "CantReadFile", new IOException());
                    java.io.InputStream credentialsStream = null;
                    credentialsStream = new FileInputStream(credentialsFile);
                    ServiceAccountCredentials serviceAccountCredentials = ServiceAccountCredentials.fromStream(credentialsStream);
                    CredentialsProvider credentialsProvider = FixedCredentialsProvider.create(serviceAccountCredentials);
                    SpeechSettings settings = SpeechSettings.newBuilder().setCredentialsProvider(credentialsProvider).build();
                } catch(java.io.FileNotFoundException e) {
                    eventSink.error("-24", "FileNotFound", e);
                } catch(IOException e) {
                    eventSink.error("-23", "CouldntReadFile", e);
                }

                try {
                    System.out.println(Class.forName("io.grpc.okhttp.OkHttpChannelProvider"));
                } catch (ClassNotFoundException e) {
                    System.out.println(e);
                }
                try {
                    System.out.println(Class.forName("io.grpc.netty.NettyChannelProvider"));
                } catch (ClassNotFoundException e) {
                    System.out.println(e);
                }
                try (SpeechClient client = SpeechClient.create(settings)) {
                    isRecording = true;

                    // Set up Google cloud streaming
                    ResponseObserver<StreamingRecognizeResponse> responseObserver =
                            new ResponseObserver<StreamingRecognizeResponse>() {
                                public void onStart(StreamController controller) {
                                }

                                public void onResponse(StreamingRecognizeResponse response) {
                                    eventSink.success(response.getResultsList().get(0));
                                }

                                public void onComplete() {
                                }

                                public void onError(Throwable t) {
                                    eventSink.error("-31", "google error", t);
                                }
                            };


                    ClientStream<StreamingRecognizeRequest> clientStream =
                            client.streamingRecognizeCallable().splitCall(responseObserver);

                    RecognitionConfig recognitionConfig =
                            RecognitionConfig.newBuilder()
                                    .setEncoding(RecognitionConfig.AudioEncoding.LINEAR16)
                                    .setLanguageCode("de-DE")
                                    .setSampleRateHertz(16000)
                                    .build();

                    StreamingRecognitionConfig streamingRecognitionConfig =
                            StreamingRecognitionConfig.newBuilder().setConfig(recognitionConfig).build();

                    StreamingRecognizeRequest request =
                            StreamingRecognizeRequest.newBuilder()
                                    .setStreamingConfig(streamingRecognitionConfig)
                                    .build(); // The first request in a streaming call has to be a config

                    clientStream.send(request);

                    // Repeatedly push audio samples to stream
                    while (record) {
                        // Read audio data into new byte array
                        byte[] audio = new byte[BUFFER_SIZE];
                        recorder.read(audio, 0, BUFFER_SIZE);

                        request =
                                StreamingRecognizeRequest.newBuilder()
                                        .setAudioContent(ByteString.copyFrom(audio))
                                        .build();
                        clientStream.send(request);
                    }
                    isRecording = false;
                } catch (Throwable e) {
                    eventSink.error("-31", "UnknownException", e);
                }
            }
        };
    }
}

